//==========================================================================
//
//      if_8139.c
//
//	RealTek 8139 ethernet driver
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//####BSDCOPYRIGHTBEGIN####
//
// -------------------------------------------
//
// Portions of this software may have been derived from OpenBSD or
// other sources, and are covered by the appropriate copyright
// disclaimers included herein.
//
// -------------------------------------------
//
//####BSDCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Eric Doenges
// Contributors:
// Date:         2003-07-09
// Purpose:
// Description:  hardware driver for RealTek 8139 ethernet
// Notes:        This is a very basic driver that will send and receive
//               packets and not much else. A lot of features of the 8139
//               are not supported (power management, MII interface,
//               access to the serial eeprom, 'twister tuning', etc.).
//
//               Many of the configuration options (like media type and
//               speed)the 8139 has are taken directly from the serial
//               eeprom and are not currently configurable from this driver.
//
//               Since the RealTek 8139 data sheet is filled with omissions,
//               half-truths and outright falsehoods, anyone attempting to
//               write a driver for this chip is more or less forced to look
//               at existing code to figure out how it works. I have looked
//               at the Linux and OpenBSD drivers to understand the basic
//               workings of the chip; however, I have not copied any code
//               from those drivers to avoid tainting eCos' license.
//        FIXME:
//
//####DESCRIPTIONEND####
//
//==========================================================================
#include <pkgconf/system.h>
#ifdef CYGPKG_IO_ETH_DRIVERS
#include <pkgconf/io_eth_drivers.h>
#endif
#include <pkgconf/devs_eth_rltk_8139.h>

#include <cyg/infra/cyg_type.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/drv_api.h>
#include <cyg/io/eth/netdev.h>
#include <cyg/io/eth/eth_drv.h>

#include <string.h> /* for memset */

#ifdef CYGPKG_IO_PCI
#include <cyg/io/pci.h>
#else
#error "This driver requires the PCI package (CYGPKG_IO_PCI)"
#endif

#include <cyg/io/pci.h>

/* Required for CYGARC_PHYSICAL_ADDRESS */
#include CYGHWR_MEMORY_LAYOUT_H

#include <cyg/devs/eth/if_8139.h>

/* Allow platform-specific configuration of the driver */
#include CYGDAT_DEVS_ETH_RLTK_8139_INL

/*
 * Ugly hack. This is necessary to get RedBoot to build, since these
 * are somehow not defined.
 */
#ifndef EINVAL
#define EINVAL -2
#endif

#ifndef ENOERR
#define ENOERR 0
#endif


/*
 * Local driver functions
 */
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
static cyg_uint32 rltk8139_isr(cyg_vector_t vector, cyg_addrword_t data);
static void rltk8139_dsr(cyg_vector_t vector, cyg_ucount32 count,
                         cyg_addrword_t data);
#endif

static void rltk8139_reset(struct eth_drv_sc *sc);
static bool rltk8139_init(struct cyg_netdevtab_entry *tab);
static void rltk8139_start(struct eth_drv_sc *sc, unsigned char *enaddr,
                           int flags);
static void rltk8139_stop(struct eth_drv_sc *sc);
static int rltk8139_control(struct eth_drv_sc *sc, unsigned long key,
                            void *data, int   data_length);
static int rltk8139_can_send(struct eth_drv_sc *sc);
static void rltk8139_send(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list,
                          int sg_len, int total_len, unsigned long key);
static void rltk8139_recv(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list,
                          int sg_len);
static void rltk8139_deliver(struct eth_drv_sc *sc);
static void rltk8139_poll(struct eth_drv_sc *sc);
static int rltk8139_int_vector(struct eth_drv_sc *sc);


/*
 * Define inline functions to access the card. This will also handle
 * endianess issues in one place. This code was lifted from the eCos
 * i82559 driver.
 */
#if (CYG_BYTEORDER == CYG_MSBFIRST)
#define HAL_CTOLE32(x)  ((((x) & 0xff) << 24) | (((x) & 0xff00) << 8) | (((x) & 0xff0000) >> 8) | (((x) >> 24) & 0xff))
#define HAL_LE32TOC(x)  ((((x) & 0xff) << 24) | (((x) & 0xff00) << 8) | (((x) & 0xff0000) >> 8) | (((x) >> 24) & 0xff))

#define HAL_CTOLE16(x)  ((((x) & 0xff) << 8) | (((x) & 0xff00) >> 8))
#define HAL_LE16TOC(x)  ((((x) & 0xff) << 8) | (((x) & 0xff00) >> 8))

static inline
void OUTB(cyg_uint8 value,
          cyg_uint32 io_address)
{
  HAL_WRITE_UINT8( io_address, value);
}

static inline void OUTW(cyg_uint16 value, cyg_uint32 io_address)
{
    HAL_WRITE_UINT16( io_address, (((value & 0xff) << 8) | ((value & 0xff00) >> 8)) );
}

static inline void OUTL(cyg_uint32 value, cyg_uint32 io_address)
{
    HAL_WRITE_UINT32( io_address,
                      ((((value) & 0xff) << 24) | (((value) & 0xff00) << 8) | (((value) & 0xff0000) >> 8) | (((value) >> 24) & 0xff)) );
}

static inline cyg_uint8 INB(cyg_uint32 io_address)
{
    cyg_uint8 d;
    HAL_READ_UINT8( io_address, d );
    return d;
}

static inline cyg_uint16 INW(cyg_uint32 io_address)
{
    cyg_uint16 d;
    HAL_READ_UINT16( io_address, d );
    return (((d & 0xff) << 8) | ((d & 0xff00) >> 8));
}

static inline cyg_uint32 INL(cyg_uint32 io_address)
{
    cyg_uint32 d;
    HAL_READ_UINT32( io_address, d );
    return ((((d) & 0xff) << 24) | (((d) & 0xff00) << 8) | (((d) & 0xff0000) >> 8) | (((d) >> 24) & 0xff));
}
#else

// Maintaining the same styleee as above...
#define HAL_CTOLE32(x)  ((((x))))
#define HAL_LE32TOC(x)  ((((x))))

#define HAL_CTOLE16(x)  ((((x))))
#define HAL_LE16TOC(x)  ((((x))))

static inline void OUTB(cyg_uint8  value, cyg_uint32 io_address)
{   HAL_WRITE_UINT8( io_address, value );   }

static inline void OUTW(cyg_uint16 value, cyg_uint32 io_address)
{   HAL_WRITE_UINT16( io_address, value );   }

static inline void OUTL(cyg_uint32 value, cyg_uint32 io_address)
{   HAL_WRITE_UINT32( io_address, value );   }

static inline cyg_uint8  INB(cyg_uint32 io_address)
{   cyg_uint8  _t_; HAL_READ_UINT8(  io_address, _t_ ); return _t_;   }

static inline cyg_uint16 INW(cyg_uint32 io_address)
{   cyg_uint16 _t_; HAL_READ_UINT16( io_address, _t_ ); return _t_;   }

static inline cyg_uint32 INL(cyg_uint32 io_address)
{   cyg_uint32 _t_; HAL_READ_UINT32( io_address, _t_ ); return _t_;   }

#endif // byteorder


/*
 * The code below will search all PCI busses for the Nth 8139 device.
 * We might want to move this into the .inl file so that it can be adapted
 * to different platforms.
 */
static cyg_bool
rltk8139_find_match_func(cyg_uint16 vendorID,
                         cyg_uint16 deviceID,
                         cyg_uint32 classID,
                         void *p)
{
  int i;


  for (i = 0; i < CYGNUM_DEVS_ETH_RLTK_8139_KNOWN_ALIASES; ++i) {
    if (((known8139_aliases[i].vendorID == PCI_ANY_ID) ||
         (known8139_aliases[i].vendorID == vendorID)) &&
        ((known8139_aliases[i].deviceID == PCI_ANY_ID) ||
         (known8139_aliases[i].deviceID == deviceID)))
      return true;
  }

  return false;
}


/*
 * Find the Nth 8139 device on all attached PCI busses and do some initial
 * PCI-type initialization. Also setup the interrupt for use in eCos.
 */
static bool
rltk8139_find(int Nth,
              Rltk8139_t *info)
{
  cyg_pci_device_id deviceID;
  cyg_pci_device deviceInfo;
  cyg_uint16 command;
  int found = -1;


  /* Go through all PCI devices until we find the Nth 8139 chip */
  do {
    deviceID = CYG_PCI_NULL_DEVID;
    if (!cyg_pci_find_matching(&rltk8139_find_match_func, NULL, &deviceID))
      return false;
    else
      found += 1;
  } while (found != Nth);

#ifdef DEBUG
  diag_printf("Found %dth RealTek 8139 device on bus %d, devfn 0x%02x:\n",
              Nth, CYG_PCI_DEV_GET_BUS(deviceID),
              CYG_PCI_DEV_GET_DEVFN(deviceID));
#endif

  /* Save device ID in driver private data in case we ever need it again */
  info->deviceID = deviceID;

  /* Now that we have found the device, we can extract some data about it */
  cyg_pci_get_device_info(deviceID, &deviceInfo);

  /* Get the assigned interrupt and set up ISR and DSR for it. */
  if (cyg_pci_translate_interrupt(&deviceInfo, &info->vector)) {
#ifdef DEBUG
    diag_printf(" Wired to HAL interrupt vector %d\n", info->vector);
#endif

#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
    cyg_drv_interrupt_create(info->vector, 0, (CYG_ADDRWORD)info,
                             rltk8139_isr,
                             rltk8139_dsr,
                             &info->interruptHandle,
                             &info->interrupt);
    cyg_drv_interrupt_attach(info->interruptHandle);
#endif
  }
  else {
#ifdef DEBUG
    diag_printf(" Does not generate interrupts.\n");
#endif
  }

  /*
   * Enable memory mapped and port based I/O and busmastering. We currently
   * only support IO space accesses; memory mapping is enabled so that bit
   * DVRLOAD in CONFIG1 is cleared automatically.
   *
   * NOTE: it seems that for some HALs (in my case, Equator's BSP HAL), the
   *       device is already activated at this point, even though eCos'
   *       documentation suggests it shouldn't be. At least in my case,
   *       this is not a problem.
   */
  cyg_pci_read_config_uint16(deviceInfo.devid, CYG_PCI_CFG_COMMAND, &command);
  command |= (CYG_PCI_CFG_COMMAND_IO | CYG_PCI_CFG_COMMAND_MEMORY |
              CYG_PCI_CFG_COMMAND_MASTER);
  cyg_pci_write_config_uint16(deviceInfo.devid, CYG_PCI_CFG_COMMAND, command);

  /*
   * This is the base address used to talk to the device. The 8139's IOAR
   * and MEMAR registers are BAR0 and BAR1, respectively.
   */
  info->baseAddress = deviceInfo.base_map[0];

  /*
   * Read the MAC address. The RealTek data sheet claims this should only be
   * read in 4 byte accesses, but that makes little sense since the MAC is
   * only 6 bytes (IDR6 and IDR7 are defined as 'reserved'). The RealTek
   * data sheet is full of garbage like this.
   */
  for (found = 0; found < 6; ++found)
    info->MAC[found] = INB(info->baseAddress + IDR0 + found);

  return true;
}


#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
/*
 * Interrupt service routine. We do not clear the interrupt status bits
 * (since this should really only be done after handling whatever caused
 * the interrupt, and that is done in the '_deliver' routine), but instead
 * clear the interrupt mask. We save the previous state of the interrupt
 * mask so that '_deliver' can restore it after completion.
 *
 * Note: with current code, we would not need to save and restore the
 *       interrupt mask, since it is only changed in two places - the ISR
 *       and the '_deliver' routine. However, saving the mask state makes
 *       us more flexible for the future.
 */
static cyg_uint32
rltk8139_isr(cyg_vector_t vector,
             cyg_addrword_t data)
{
  Rltk8139_t *info;
  cyg_uint16 isr;


  info = (Rltk8139_t *)data;

  /*
   * First, check if the interrupt is really for us in case we are
   * sharing interrupts with another device. The RealTek data sheet claims
   * that reading the interrupt state register will clear all it's bits;
   * this is complete bullshit.
   */
  if (!(isr = INW(info->baseAddress + ISR)))
    return 0;

  /*
   * Save the current mask value and then clear it to remove the
   * interrupt request.
   */
  info->interruptMask = INW(info->baseAddress + IMR);
  OUTW(0, info->baseAddress + IMR);

  return (CYG_ISR_HANDLED | CYG_ISR_CALL_DSR);
}


/*
 * Delayed service routine.
 */
static void
rltk8139_dsr(cyg_vector_t vector,
             cyg_ucount32 count,
             cyg_addrword_t data)
{
  Rltk8139_t *info;
  struct eth_drv_sc *sc;


  info = (Rltk8139_t *)data;
  sc = info->netDeviceTableEntry->device_instance;

  /* Call the generic ethernet support DSR */
  eth_drv_dsr(vector, count, (cyg_addrword_t)sc);
}
#endif


/*
 * Reset the chip. This function is not exported to higher level software.
 */
static void
rltk8139_reset(struct eth_drv_sc *sc)
{
  rltk8139_stop(sc);
  rltk8139_start(sc, NULL, 0);
}


/*
 * Initialize the network interface
 */
static bool
rltk8139_init(struct cyg_netdevtab_entry *tab)
{
  struct eth_drv_sc *sc;
  Rltk8139_t *info;


  sc = (struct eth_drv_sc *)(tab->device_instance);
  info = (Rltk8139_t *)(sc->driver_private);

  /* Link private data back to the net device table entry */
  info->netDeviceTableEntry = tab;

  /*
   * Initialize the eCos PCI library. According to the documentation, it
   * is safe to call this function multiple times, so we call it just to
   * be sure it has been done.
   */
  cyg_pci_init();

  /*
   * Scan the PCI bus for the specified chip. The '_find' function will also
   * do some basic PCI initialization.
   */
  if (!rltk8139_find(info->deviceNum, info)) {
#ifdef DEBUG
    diag_printf("Could not find RealTek 8139 chip #%d.\n", info->deviceNum);
#endif
    return false;
  }

  /*
   * The initial tx threshold is set here to prevent it from being reset
   * with every _start().
   */
  info->txThreshold = 3;

  /* Initialize upper level driver */
  (sc->funs->eth_drv->init)(sc, info->MAC);

  return true;
}


/*
 * (Re)Start the chip, initializing data structures and enabling the
 * transmitter and receiver. Currently, 'flags' is unused.
 */
static void
rltk8139_start(struct eth_drv_sc *sc,
               unsigned char *enaddr,
               int flags)
{
  Rltk8139_t *info;
  int i;


#ifdef DEBUG
  diag_printf("rltk8139_start(%s)\n", sc->dev_name);
#endif

  info = (Rltk8139_t *)(sc->driver_private);

  /*
   * Reset the chip. Existing driver code implies that this may take up
   * to 10ms; since I don't know under what exact circumstances this code may
   * be called I busy wait here regardless.
   */
  OUTB(RST, info->baseAddress + CR);
  while (INB(info->baseAddress + CR) & RST);

#ifdef DEBUG
  diag_printf("8139 was successfully reset.\n");
#endif

  /*
   * Clear the key storage area. These keys are used by the eCos networking
   * support code to keep track of individual packets.
   */
  for (i = 0; i < NUM_TX_DESC; ++i)
    info->txDescKey[i] = 0;

  /* Initialize transmission buffer control */
  info->txFreeDesc = 0;
  info->txNumFreeDesc = NUM_TX_DESC;

  /*
   * Set the requested MAC address if enaddr != NULL. This is a special
   * feature of my '_start' function since it's used to reset the chip after
   * errors as well.
   */
  if (enaddr != NULL) {
    for (i = 0; i < 6; ++i) {
      info->MAC[i] = enaddr[i];
      OUTB(enaddr[i], info->baseAddress + IDR0 + i);
    }
  }

  /*
   * Now setup the transmission and reception buffers. These could be done
   * in _init() and kept around, but putting them here fits better logically.
   * Note the use of CYGARC_PHYSICAL_ADDRESS for platforms on which virtual
   * and physical addresses do not match.
   */
  OUTL(CYGARC_PHYSICAL_ADDRESS((cyg_uint32)(info->txBuffer + 0 * TX_BUF_SIZE)),
       info->baseAddress + TSAD0);
  OUTL(CYGARC_PHYSICAL_ADDRESS((cyg_uint32)(info->txBuffer + 1 * TX_BUF_SIZE)),
       info->baseAddress + TSAD1);
  OUTL(CYGARC_PHYSICAL_ADDRESS((cyg_uint32)(info->txBuffer + 2 * TX_BUF_SIZE)),
       info->baseAddress + TSAD2);
  OUTL(CYGARC_PHYSICAL_ADDRESS((cyg_uint32)(info->txBuffer + 3 * TX_BUF_SIZE)),
       info->baseAddress + TSAD3);
  OUTL(CYGARC_PHYSICAL_ADDRESS((cyg_uint32)info->rxRing),
       info->baseAddress + RBSTART);

  /* Enable the transmitter and receiver */
  OUTB(INB(info->baseAddress + CR) | (TE|RE), info->baseAddress + CR);

  /* Clear the missed packet counter */
  OUTL(0, info->baseAddress + MPC);

  /*
   * It seems the receiver and transmitter configuration can only
   * be set after the transmitter/receiver have been enabled. Of course,
   * this is not documented in the data sheet, but we've come to expect
   * that of RealTek.
   */
  OUTL(TXCFG, info->baseAddress + TCR);
  OUTL(RXCFG | AM, info->baseAddress + RCR);

  /*
   * Enable the transmitter and receiver again. I'm not sure why this is
   * necessary; the Linux driver does it.
   */
  OUTB(INB(info->baseAddress + CR) | (TE|RE), info->baseAddress + CR);

  /*
   * Enable interrupts. We do not currently handle all types of errors.
   */
  info->interruptMask = IR_SERR|IR_FOVW|IR_RXOVW|IR_TER|IR_TOK|IR_RER|IR_ROK;
  OUTW(info->interruptMask, info->baseAddress + IMR);
}


/*
 * Stop the chip, disabling the transmitter and receiver.
 */
static void
rltk8139_stop(struct eth_drv_sc *sc)
{
  Rltk8139_t *info;


#ifdef DEBUG
  diag_printf("rltk8139_stop(%s)\n", sc->dev_name);
#endif

  info = (Rltk8139_t *)(sc->driver_private);

  /* Disable receiver and transmitter */
  OUTB(INB(info->baseAddress + CR) & ~(TE|RE), info->baseAddress + CR);

  /* Mask all interrupts */
  info->interruptMask = 0;
  OUTW(info->interruptMask, info->baseAddress + IMR);
}


/*
 * 8139 ioctl
 */
static int
rltk8139_control(struct eth_drv_sc *sc,
                 unsigned long key,
                 void *data,
                 int data_length)
{
  int i;
  Rltk8139_t *info;


#ifdef DEBUG
  diag_printf("rltk8139_control(%08x, %lx)\n", sc, key);
#endif

  info = (Rltk8139_t *)(sc->driver_private);

  switch (key) {
#ifdef ETH_DRV_SET_MAC_ADDRESS
  case ETH_DRV_SET_MAC_ADDRESS:
    if ( 6 != data_length )
      return -EINVAL;

    /* Set the mac address */
    for (i = 0; i < 6; ++i) {
      info->MAC[i] = *(((cyg_uint8 *)data) + i);
      OUTB(info->MAC[i], info->baseAddress + IDR0 + i);
    }
    return ENOERR;
#endif

#ifdef ETH_DRV_GET_MAC_ADDRESS
    case ETH_DRV_GET_MAC_ADDRESS:
      if (6 != data_length)
        return -EINVAL;

      memcpy(data, info->MAC, 6);
      return ENOERR;
#endif

#ifdef ETH_DRV_GET_IF_STATS_UD
    case ETH_DRV_GET_IF_STATS_UD: // UD == UPDATE
      //ETH_STATS_INIT( sc );    // so UPDATE the statistics structure
#endif
      // drop through
#ifdef ETH_DRV_GET_IF_STATS
  case ETH_DRV_GET_IF_STATS:
#endif
#if defined(ETH_DRV_GET_IF_STATS) || defined (ETH_DRV_GET_IF_STATS_UD)
    break;
#endif

#ifdef ETH_DRV_SET_MC_LIST
  case ETH_DRV_SET_MC_LIST:
    /*
     * Clear the AM bit in receive config. We should also configure
     * the multicast registers, but I don't know how this works on the
     * 8139.
     */
    OUTL(INL(info->baseAddress + RCR) & ~AM, info->baseAddress + RCR);
    return ENOERR;
#endif // ETH_DRV_SET_MC_LIST

#ifdef ETH_DRV_SET_MC_ALL
  case ETH_DRV_SET_MC_ALL:
    /* Set the AM bit in receive config */
    OUTL(INL(info->baseAddress + RCR) | AM, info->baseAddress + RCR);
    return ENOERR;
#endif // ETH_DRV_SET_MC_ALL

  default:
    return EINVAL;
  }

  return EINVAL;
}


/*
 * Check if a new packet can be sent.
 */
static int
rltk8139_can_send(struct eth_drv_sc *sc)
{
  return ((Rltk8139_t *)(sc->driver_private))->txNumFreeDesc;
}


/*
 * Send a packet over the wire.
 */
static void
rltk8139_send(struct eth_drv_sc *sc,
              struct eth_drv_sg *sg_list,
              int sg_len,
              int total_len,
              unsigned long key)
{
  Rltk8139_t *info;
  cyg_uint8 *tx_buffer;
  struct eth_drv_sg *last_sg;
  int desc;


  info = (Rltk8139_t *)(sc->driver_private);

#ifdef DEBUG
  diag_printf("rltk8139_send(%s, %08x, %d, %d, %08lx)\n",
              sc->dev_name, sg_list, sg_len, total_len, key);
#endif

  CYG_ASSERT(total_len <= TX_BUF_SIZE, "packet too long");


  /*
   * Get the next free descriptor to send. We lock out all interrupts
   * and scheduling because we really, really do not want to be interrupted
   * at this point.
   *
   * IMPORTANT NOTE: the RealTek data sheet does not really make this clear,
   * but when they talk about a 'ring' of transmit descriptors, they
   * _really_ mean it, i.e. you _must_ use descriptor #1 after descriptor
   * #0 even if transmission of descriptor #0 has already completed. Why
   * they didn't implement this in hardware then is anyone's guess.
   */
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
  cyg_interrupt_disable();
#endif

  /*
   * Sanity check to see if '_send' was called even though there is no free
   * descriptor. This is probably unnecessary.
   */
  if (info->txNumFreeDesc == 0) {
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
    cyg_interrupt_enable();
#endif
#ifdef DEBUG
    diag_printf("No descriptor free but send called regardless ?\n");
#endif
    return;
  }

  /*
   * Get the free descriptor and advance the descriptor counter modulo
   * TX_NUM_DESC. We assume that TX_NUM_DESC will always be a power of 2.
   */
  desc = info->txFreeDesc;
  info->txFreeDesc = (info->txFreeDesc + 1) & (NUM_TX_DESC - 1);

  /* Decrement the number of free descriptors */
  info->txNumFreeDesc -= 1;

#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
  /* Reenable interrupts at this point */
  cyg_interrupt_enable();
#endif

  /*
   * Determine the buffer memory to use and tell the hardware about it.
   * Since we use fixed buffer addresses, we do not need to set up TSADx.
   * Memorize the key so we can call the tx_done callback correctly.
   *
   * While it would be possible to set TSADx to the packet directly if
   * it is stored in a linear memory area with 32 bit alignment, it seems
   * this happens so seldomly that it's simply not worth the extra
   * runtime check.
   */
  tx_buffer = info->txBuffer + TX_BUF_SIZE * desc;
  info->txDescKey[desc] = key;

  /*
   * Copy the data to the designated position. Note that unlike the eCos
   * Intel 82559 driver, we simply assume that all the scatter/gather list
   * elements' lengths will add up to total_len exactly.
   */
  for (last_sg = &sg_list[sg_len]; sg_list < last_sg; ++sg_list) {
    memcpy(tx_buffer, (void *)(sg_list->buf), sg_list->len);
    tx_buffer += sg_list->len;
  }

  /*
   * Make sure the packet has the minimum ethernet packet size, padding
   * with zeros if necessary.
   */
  if (total_len < MIN_ETH_FRAME_SIZE) {
    memset(tx_buffer, 0, MIN_ETH_FRAME_SIZE - total_len);
    total_len = MIN_ETH_FRAME_SIZE;
  }

  /*
   * Now setup the correct transmit descriptor to actually send the data.
   * The early TX threshold is incremented by the driver until we reach a
   * size that prevents transmit underruns.
   */
  OUTL((info->txThreshold << ERTXTH_SHIFT) | (total_len & SIZE),
       info->baseAddress + TSD0 + (desc<<2));
}


/*
 * This routine actually retrieves data from the receive ring by
 * copying it into the specified scatter/gather buffers.
 */
static void
rltk8139_recv(struct eth_drv_sc *sc,
              struct eth_drv_sg *sg_list,
              int sg_len)
{
  Rltk8139_t *info;
  struct eth_drv_sg *last_sg;
  cyg_uint8 *rx_buffer;


  info = (Rltk8139_t *)(sc->driver_private);
  rx_buffer = info->rxCurrent;
  for (last_sg = &sg_list[sg_len]; sg_list < last_sg; ++sg_list) {
    memcpy((void *)(sg_list->buf), rx_buffer, sg_list->len);
    rx_buffer += sg_list->len;
  }
}


/*
 * This function does all the heavy lifting associated with interrupts.
 */
static void
rltk8139_deliver(struct eth_drv_sc *sc)
{
  Rltk8139_t *info;
  cyg_uint16 status, pci_status;
  cyg_uint32 tsd;
  int desc;


  info = (Rltk8139_t *)(sc->driver_private);

  /*
   * The RealTek data sheet claims that reading the ISR will clear
   * it. This is complete bullshit; to clear a bit in the ISR, a
   * '1' must be written to the ISR instead. We immediately clear
   * the interrupt bits at this point.
   */
  status = INW(info->baseAddress + ISR);
  OUTW(status, info->baseAddress + ISR);

#ifdef DEBUG
  diag_printf("rltk8139_deliver(%s): %04x\n", sc->dev_name, status);
#endif

  /*
   * Check for a PCI error. This seems like a very serious error to
   * me, so we will reset the chip and hope for the best.
   */
  if (status & IR_SERR) {
    cyg_pci_read_config_uint16(info->deviceID, 0x6, &pci_status);
    cyg_pci_write_config_uint16(info->deviceID, 0x6, pci_status);
#ifdef DEBUG
    diag_printf("%s: PCI error: %04x\n", sc->dev_name, pci_status);
#endif

    rltk8139_reset(sc);
    return;
  }

  /* Check for transmission complete (with errors or not) */
  if ((status & IR_TER) || (status & IR_TOK)) {
    /*
     * Figure out which descriptors' status must be checked. We lock out
     * interrupts while manipulating the descriptor list because we do not
     * want to be interrupted at this point.
     */
    while (1) {
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
      cyg_interrupt_disable();
#endif

      /* Check if all descriptors are ready, in which case we are done. */
      if (info->txNumFreeDesc >= NUM_TX_DESC) {
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
        cyg_interrupt_enable();
#endif
        break;
      }

      desc = (info->txFreeDesc - (NUM_TX_DESC - info->txNumFreeDesc)) &
        (NUM_TX_DESC - 1);
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
      cyg_interrupt_enable();
#endif
      /* Get the current status of the descriptor */
      tsd = INL(info->baseAddress + TSD0 + (desc<<2));

      /*
       * If a transmit FIFO underrun occurred, increment the threshold
       * value.
       */
      if ((tsd & TUN) && info->txThreshold < 64)
        info->txThreshold += 1;

      /*
       * Check if a transmission completed OK. RealTek's data sheet implies
       * that a successful transmission that experiences underrun will only
       * set TUN. This is not true; TOK is set for all successful
       * transmissions.
       */
      if (tsd & TOK) {
        (sc->funs->eth_drv->tx_done)(sc, info->txDescKey[desc], 0);
      }
      else if (tsd & TABT) {
        /*
         * Set the CLRABT bit in TCR. Since I haven't encountered any
         * transmission aborts so far, I don't really know if this code
         * will work or not.
         */
        OUTL(INL(info->baseAddress + TCR) & CLRABT, info->baseAddress + TCR);
        (sc->funs->eth_drv->tx_done)(sc, info->txDescKey[desc], -1);
      }
      else {
        /*
         * This descriptor is not ready. Since the descriptors are used
         * in a ring, this means that no more descriptors are ready.
         */
        break;
      }

      /*
       * Clear the saved key value. This is not really necessary, since
       * the key value is never used to determine if a descriptor is
       * valid or not. However, clearing it is a tidier IMO.
       */
      info->txDescKey[desc] = 0;

      /*
       * Increment the free descriptor count and go through the loop again
       * to see if more descriptors are ready.
       */
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
      cyg_interrupt_disable();
#endif
      info->txNumFreeDesc += 1;
#ifndef CYGPKG_IO_ETH_DRIVERS_STAND_ALONE
      cyg_interrupt_enable();
#endif
    }
  }

  if (status & IR_ROK) {
    /*
     * Received a frame. Note that '_deliver' does not actually copy any
     * data; it just determines how many bytes are available and tells
     * the generic ethernet driver about it, which then calls into
     * the '_recv' function to actually copy the data.
     */
    cyg_uint16 rxPos;
    cyg_uint32 header;
    int length;

    /*
     * CAPR contains the index into the receive buffer. It is controlled
     * completely in software. For some reason, CAPR points 16 bytes
     * before the actual start of the packet.
     */
    rxPos = (INW(info->baseAddress + CAPR) + 16) % RX_BUF_LEN;

    /*
     * Loop until the rx buffer is empty. I have no idea how the 8139
     * determines if the buffer still contains a packet; it may check
     * that CAPR points 16 bytes before CBR.
     */
    while (!(INB(info->baseAddress + CR) & BUFE)) {
      /*
       * The 8139 prepends each packet with a 32 bit packet header that
       * contains a 16 bit length and 16 bit status field, in little-endian
       * byte order. (The RealTek data sheet does document the status field,
       * but not the complete header).
       */
      header = HAL_LE32TOC(*((cyg_uint32 *)&info->rxRing[rxPos]));

      /*
       * If the 8139 is still copying data for this packet into the
       * receive ring, it will set packet length to 0xfff0. This shouldn't
       * ever happen because we do not use early receive.
       */
      if ((header >> 16) == 0xFFF0)
        break;

      /*
       * Since the 8139 appends the ethernet CRC to every packet, we
       * must subtract 4 from the length to get the true packet length.
       */
      length = (header >> 16) - 4;

      /*
       * Check if the packet was received correctly. The OpenBSD driver
       * resets the chip if this is not the case; we attempt to salvage
       * following packets by doing nothing.
       */
      if (!(header & HDR_ROK)) {
      }
      else {
        /*
         * Packet was received OK. Determine from where to start copying
         * bytes. This is saved in the driver private area so rlt8139_recv
         * doesn't have to redetermine this information. Then, inform
         * the generic ethernet driver about the packet.
         */
        info->rxCurrent = info->rxRing + rxPos + 4;

        /* Tell eCos about the packet */
        (sc->funs->eth_drv->recv)(sc, length);
      }

      /*
       * Update CAPR. CAPR must be aligned to a 32 bit boundary, and should
       * point 16 bytes before it's actual position (don't ask !).
       */
      rxPos = ((rxPos + length + 8 + 3) & ~3) % RX_BUF_LEN;
      OUTW((rxPos - 16) % RX_BUF_LEN, info->baseAddress + CAPR);
    }
  }

  if (status & IR_RXOVW) {
    /*
     * In case of a receive buffer overflow, the RealTek data sheet claims we
     * should update CAPR and then write a '1'to ROK in ISR. However, none of
     * the other 8139 drivers I have looked at do this, so we will just reset
     * the chip instead.
     */
#ifdef DEBUG
    diag_printf("receive buffer overflow\n");
#endif
    rltk8139_reset(sc);
    return;
  }

  if (status & IR_FOVW) {
    /*
     * Rx FIFO overflow. According to RealTek's data sheet, this is cleared
     * by writing a '1' to RXOVW. Again, none of the 8139 drivers I have
     * seen actually do this, so we reset the chip instead.
     */
#ifdef DEBUG
    diag_printf("receive FIFO overflow\n");
#endif
    rltk8139_reset(sc);
    return;
  }

  /* Finally, restore the state of the interrupt mask register */
  OUTW(info->interruptMask, info->baseAddress + IMR);
}


/*
 * The way I understand the eCos docs, '_poll' is periodically called
 * when the ethernet driver is operated in non-interrupt (standalone) mode,
 * for instance in redboot.
 */
static void
rltk8139_poll(struct eth_drv_sc *sc)
{
  Rltk8139_t *info;
  cyg_uint16 isr;


#ifdef DEBUG
  diag_printf("rltk8139_poll(%s)\n", sc->dev_name);
#endif

  info = (Rltk8139_t *)(sc->driver_private);

  /*
   * Get the current interrupt status. If anything changed, call
   * _deliver.
   */
  if (!(isr = INW(info->baseAddress + ISR)))
    return;

  /*
   * Save the current mask value and then clear it. This is not necessary
   * if we could guarantee this function will never be called if interrupts
   * are not used.
   */
  info->interruptMask = INW(info->baseAddress + IMR);
  OUTW(0, info->baseAddress + IMR);

  rltk8139_deliver(sc);
}


/*
 * Return the interrupt vector used by this device.
 */
static int
rltk8139_int_vector(struct eth_drv_sc *sc)
{
  return ((Rltk8139_t *)(sc->driver_private))->vector;
}
